package com.framwork.util;


import java.beans.*;
import java.awt.Color;
import java.util.*;
import java.lang.reflect.*;

/**
 * A utility class that provides methods that are useful for dealing with
 * Java Beans.
 * @author BILL
 * @version $Revision: 1.0 $ $Date: 2004-8-11 11:31 $
 */

public class BeanUtils {

    /**
     * Sets the properties of a Java Bean based on the String name/value pairs in
     * the specifieed Map. Because this method has to know how to convert a
     * String value into the correct type for the bean, only a few bean property
     * types are supported. They are: String, boolean, int, long, float, double,
     * Color, and Class.<p>
     *
     * If key/value pairs exist in the Map that don't correspond to properties
     * of the bean, they will be ignored.
     *
     * @param bean the JavaBean to set properties on.
     * @param properties String name/value pairs of the properties to set.
     */
    public static void setProperties(Object bean, Map properties) {
        try {
            // Loop through all the property names in the Map
            for (Iterator iter = properties.keySet().iterator(); iter.hasNext();) {
                String propName = (String)iter.next();
                try {
                    // Create a property descriptor for the named property. If
                    // the bean doesn't have the named property, an
                    // Introspection will be thrown.
                    PropertyDescriptor descriptor = new PropertyDescriptor(
                            propName, bean.getClass());
                    // Load the class type of the property.
                    Class propertyType = descriptor.getPropertyType();
                    // Get the value of the property by converting it from a
                    // String to the correct object type.
                    Object value = decode(propertyType, (String)properties.get(propName));
                    // Set the value of the bean.
                    descriptor.getWriteMethod().invoke(bean, new Object[] { value });
                }
                catch (IntrospectionException ie) {
                    // Ignore. This exception means that the key in the map
                    // does not correspond to a property of the bean.
                }
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Gets the properties from a Java Bean and returns them in a Map of String
     * name/value pairs. Because this method has to know how to convert a
     * bean property into a String value, only a few bean property
     * types are supported. They are: String, boolean, int, long, float, double,
     * Color, and Class.
     *
     * @param bean a Java Bean to get properties from.
     * @return a Map of all properties as String name/value pairs.
     */
    public static Map getProperties(Object bean) {
        Map properties = new HashMap();
        try {
            BeanInfo beanInfo = Introspector.getBeanInfo(bean.getClass());
            // Loop through all properties of the bean.
            PropertyDescriptor[] descriptors = beanInfo.getPropertyDescriptors();
            String [] names = new String[descriptors.length];
            for (int i=0; i<names.length; i++) {
                // Determine the property name.
                String name = descriptors[i].getName();
                Class type = descriptors[i].getPropertyType();
                // Decode the property value using the property type and
                // encoded String value.
                Object value = descriptors[i].getReadMethod().invoke(bean, null);
                // Add to Map, encoding the value as a String.
                properties.put(name, encode(value));
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return properties;
    }

    /**
     * Returns the PropertyDescriptor array for the specified Java Bean Class.
     * The method also does a special check to see of the bean has a BeanInfo
     * class that extends the JiveBeanInfo class. If yes, we load the
     * PropertyDescriptor array directly from that BeanInfo class rather than
     * through the Introspector in order to preserve the desired ordering of
     * properties.
     *
     * @param beanClass the Class of the JavaBean.
     * @return the PropertyDescriptor array for the specified Java Bean Class.
     */
    public static PropertyDescriptor[] getPropertyDescriptors(Class beanClass)
            throws IntrospectionException
    {
        // See if the Java Bean has a BeanInfo class that implements
        // JiveBeanInfo. If so, return the PropertyDescriptor from that
        // class. This will bypass properties of parent classes, but this is
        // the normal behavior of classes that implement JiveBeanInfo.
        try {
            BeanInfo beanInfo = (BeanInfo)Class.forName(
                    beanClass.getName() + "BeanInfo").newInstance();
            return beanInfo.getPropertyDescriptors();
        }
        catch (Exception e) { }
        // Otherwise, return the PropertyDescriptors from the Introspector.
        return Introspector.getBeanInfo(beanClass).getPropertyDescriptors();
    }

    /**
     * Encodes a bean property value as a String. If the object type is not
     * supported, null will be returned.
     *
     * @param value an Object to encode in a String representation.
     */
    private static String encode(Object value) {
        if (value instanceof String) {
            return (String)value;
        }
        if (value instanceof Boolean ||
            value instanceof Integer ||
            value instanceof Long ||
            value instanceof Float ||
            value instanceof Double)
        {
            return value.toString();
        }
        if (value instanceof Color) {
            Color color = (Color)value;
            return color.getRed() +","+ color.getGreen() +","+ color.getBlue();
        }
        if (value instanceof Class) {
            return ((Class)value).getName();
        }
        return null;
    }

    /**
     * Decodes a String into an object of the specified type. If the object
     * type is not supported, null will be returned.
     *
     * @param type the type of the property.
     * @param value the encode String value to decode.
     * @return the String value decoded into the specified type.
     */
    private static Object decode(Class type, String value) throws Exception {
        if (type.getName().equals("java.lang.String")) {
            return value;
        }
        if (type.getName().equals("boolean")) {
            return Boolean.valueOf(value);
        }
        if (type.getName().equals("int")) {
            return Integer.valueOf(value);
        }
        if (type.getName().equals("long")) {
            return Long.valueOf(value);
        }
        if (type.getName().equals("float")) {
            return Float.valueOf(value);
        }
        if (type.getName().equals("double")) {
            return Double.valueOf(value);
        }
        if (type.getName().equals("java.awt.Color")) {
            StringTokenizer tokens = new StringTokenizer(value, ",");
            int red = Integer.parseInt(tokens.nextToken());
            int green = Integer.parseInt(tokens.nextToken());
            int blue = Integer.parseInt(tokens.nextToken());
            return new Color(red, green, blue);
        }
        if (type.getName().equals("java.lang.Class")) {
            return Class.forName(value);
        }
        return null;
    }

    // This class is not instantiable.
    private BeanUtils() {
        // do nothing.
    }



    public static String getLowerToProperty(Object bean, String name) {
            Field[] fields = bean.getClass().getDeclaredFields();

            for (int j = 0; j < fields.length; j++) {
                if (fields[j].getName().equalsIgnoreCase(name))
                    return fields[j].getName();
            }

            return name;
        }

        public static Object getProperty(Object bean, String name)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {


             return getProperty(bean, name,false);
        }

         public static Object getProperty(Object bean, String name,boolean isLowerUpper)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
             if(!isLowerUpper)
             name = getLowerToProperty(bean, name);
             return getNestedProperty(bean, name);
        }

        public static Object getNestedProperty(Object bean, String name)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            do {
                int delim = name.indexOf(46);
                if (delim < 0)
                    break;
                String next = name.substring(0, delim);
                if (next.indexOf(91) >= 0)
                    bean = getIndexedProperty(bean, next);
                else
                    bean = getSimpleProperty(bean, next);
                if (bean == null)
                    throw new IllegalArgumentException("Null property value for '" + name.substring(0, delim) + "'");
                name = name.substring(delim + 1);
            } while (true);
            if (name.indexOf(91) >= 0)
                return getIndexedProperty(bean, name);
            else
                return getSimpleProperty(bean, name);
        }


        public static Object getIndexedProperty(Object bean, String name)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            int delim = name.indexOf(91);
            int delim2 = name.indexOf(93);
            if (delim < 0 || delim2 <= delim)
                throw new IllegalArgumentException("Invalid indexed property '" + name + "'");
            int index = -1;
            try {
                String subscript = name.substring(delim + 1, delim2);
                index = Integer.parseInt(subscript);
            } catch (NumberFormatException numberformatexception) {
                throw new IllegalArgumentException("Invalid indexed property '" + name + "'");
            }
            name = name.substring(0, delim);

            return getIndexedProperty(bean, name, index);
        }


        public static Class getParameterTypes(Object bean, String name) throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            name = getLowerToProperty(bean, name);
            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);
            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");
            Method readMethod = null;

            if (descriptor instanceof IndexedPropertyDescriptor)
                readMethod = ((IndexedPropertyDescriptor) descriptor).getIndexedReadMethod();
            else
                readMethod = getReadMethod(descriptor);
            if (readMethod == null) {
                throw new NoSuchMethodException("Property '" + name + "' has no getter method");
            } else {
                return readMethod.getReturnType();
            }
        }


        public static Class getParameterTypesOfWrite(Object bean, String name) throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            name = getLowerToProperty(bean, name);
            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);
            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");
            Method writeMethod = null;

            if (descriptor instanceof IndexedPropertyDescriptor)
                writeMethod = ((IndexedPropertyDescriptor) descriptor).getIndexedWriteMethod();
            else
                writeMethod = getAccessibleMethod(descriptor.getWriteMethod());
            if (writeMethod == null) {
                throw new NoSuchMethodException("Property '" + name + "' has no getter method");
            } else {
                return writeMethod.getReturnType();
            }
        }

        public static Object getIndexedProperty(Object bean, String name, int index)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);
            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");
            Method readMethod;
            if (descriptor instanceof IndexedPropertyDescriptor) {
                readMethod = ((IndexedPropertyDescriptor) descriptor).getIndexedReadMethod();
                if (readMethod != null) {
                    Object subscript[] = new Object[1];
                    subscript[0] = new Integer(index);
                    return readMethod.invoke(bean, subscript);
                }
            }
            readMethod = getReadMethod(descriptor);
            if (readMethod == null)
                throw new NoSuchMethodException("Property '" + name + "' has no getter method");
            Object value = readMethod.invoke(bean, new Object[0]);
            if (!value.getClass().isArray())
                throw new IllegalArgumentException("Property '" + name + "' is not indexed");
            else
                return Array.get(value, index);
        }


        public static Object getSimpleProperty(Object bean, String name)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            if (name.indexOf(46) >= 0)
                throw new IllegalArgumentException("Nested property names are not allowed");
            if (name.indexOf(91) >= 0)
                throw new IllegalArgumentException("Indexed property names are not allowed");
            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);
            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");
            Method readMethod = getReadMethod(descriptor);
            if (readMethod == null) {
                throw new NoSuchMethodException("Property '" + name + "' has no getter method");
            } else {
                Object value = readMethod.invoke(bean, new Object[0]);
                return value;
            }
        }

        public static PropertyDescriptor getPropertyDescriptor(Object bean, String name)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            do {
                int period = name.indexOf(46);
                if (period < 0)
                    break;
                String next = name.substring(0, period);
                if (next.indexOf(91) >= 0)
                    bean = getIndexedProperty(bean, next);
                else
                    bean = getSimpleProperty(bean, next);
                if (bean == null)
                    throw new IllegalArgumentException("Null property value for '" + name.substring(0, period) + "'");
                name = name.substring(period + 1);
            } while (true);
            int left = name.indexOf(91);
            if (left >= 0)
                name = name.substring(0, left);
            if (bean == null || name == null)
                return null;

            PropertyDescriptor descriptors[] = getPropertyDescriptors(bean);

            if (descriptors == null)
                return null;
            for (int i = 0; i < descriptors.length; i++)
                if (name.equals(descriptors[i].getName()))
                    return descriptors[i];

            return null;
        }

        public static Method getReadMethod(PropertyDescriptor descriptor) {
            return getAccessibleMethod(descriptor.getReadMethod());
        }


        public static Method getAccessibleMethod(Method method) {
            if (method == null)
                return null;
            if (!Modifier.isPublic(method.getModifiers()))
                return null;
            Class clazz = method.getDeclaringClass();
            if (Modifier.isPublic(clazz.getModifiers())) {
                return method;
            } else {
                String methodName = method.getName();
                Class parameterTypes[] = method.getParameterTypes();
                method = getAccessibleMethodFromInterfaceNest(clazz, method.getName(), method.getParameterTypes());
                return method;
            }
        }

        public static Method getAccessibleMethodFromInterfaceNest(Class clazz, String methodName, Class parameterTypes[]) {
            Method method = null;
            Class interfaces[] = clazz.getInterfaces();
            for (int i = 0; i < interfaces.length; i++) {
                if (!Modifier.isPublic(interfaces[i].getModifiers()))
                    continue;
                try {
                    method = interfaces[i].getDeclaredMethod(methodName, parameterTypes);
                } catch (NoSuchMethodException nosuchmethodexception) {
                }
                if (method != null)
                    break;
                method = getAccessibleMethodFromInterfaceNest(interfaces[i], methodName, parameterTypes);
                if (method != null)
                    break;
            }

            return method;
        }

        public static PropertyDescriptor[] getPropertyDescriptors(Object bean) {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");

            String beanClassName = bean.getClass().getName();

            PropertyDescriptor descriptors[] = null;

            BeanInfo beanInfo = null;
            try {
                beanInfo = Introspector.getBeanInfo(bean.getClass());
            } catch (IntrospectionException introspectionexception) {
                return new PropertyDescriptor[0];
            }
            descriptors = beanInfo.getPropertyDescriptors();
            if (descriptors == null)
                descriptors = new PropertyDescriptor[0];

            return descriptors;
        }

        public static void setIndexedProperty(Object bean, String name, Object value)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            int delim = name.indexOf(91);
            int delim2 = name.indexOf(93);
            if (delim < 0 || delim2 <= delim)
                throw new IllegalArgumentException("Invalid indexed property '" + name + "'");
            int index = -1;
            try {
                String subscript = name.substring(delim + 1, delim2);
                index = Integer.parseInt(subscript);
            } catch (NumberFormatException numberformatexception) {
                throw new IllegalArgumentException("Invalid indexed property '" + name + "'");
            }
            name = name.substring(0, delim);
            setIndexedProperty(bean, name, index, value);
        }

        public static void setIndexedProperty(Object bean, String name, int index, Object value)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);
            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");
            if (descriptor instanceof IndexedPropertyDescriptor) {
                Method writeMethod = ((IndexedPropertyDescriptor) descriptor).getIndexedWriteMethod();
                if (writeMethod != null) {
                    Object subscript[] = new Object[2];
                    subscript[0] = new Integer(index);
                    subscript[1] = value;
                    writeMethod.invoke(bean, subscript);
                    return;
                }
            }
            Method readMethod = descriptor.getReadMethod();
            if (readMethod == null)
                throw new NoSuchMethodException("Property '" + name + "' has no getter method");
            Object array = readMethod.invoke(bean, new Object[0]);
            if (!array.getClass().isArray()) {
                throw new IllegalArgumentException("Property '" + name + "' is not indexed");
            } else {
                Array.set(array, index, value);
                return;
            }
        }

        public static void setNestedProperty(Object bean, String name, Object value)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            do {
                int delim = name.indexOf(46);
                if (delim < 0)
                    break;
                String next = name.substring(0, delim);

                if (next.indexOf(91) >= 0)
                    bean = getIndexedProperty(bean, next);
                else
                    bean = getSimpleProperty(bean, next);
                if (bean == null)
                    throw new IllegalArgumentException("Null property value for '" + name.substring(0, delim) + "'");
                name = name.substring(delim + 1);
            } while (true);
            if (name.indexOf(91) >= 0)
                setIndexedProperty(bean, name, value);
            else
                setSimpleProperty(bean, name, value);
        }

       public static void setProperty(Object bean, String name, Object value)
         throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
             setProperty(bean,name,value,false);
       }

        public static void setProperty(Object bean, String name, Object value,boolean isLowerUpper)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            if(!isLowerUpper)
            name = getLowerToProperty(bean, name);

            setNestedProperty(bean, name, value);
        }

        public static void setSimpleProperty(Object bean, String name, Object value)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {


            if (bean == null)
                throw new IllegalArgumentException("No bean specified");
            if (name == null)
                throw new IllegalArgumentException("No name specified");
            if (name.indexOf(46) >= 0)
                throw new IllegalArgumentException("Nested property names are not allowed");
            if (name.indexOf(91) >= 0)
                throw new IllegalArgumentException("Indexed property names are not allowed");

            PropertyDescriptor descriptor = getPropertyDescriptor(bean, name);


            if (descriptor == null)
                throw new NoSuchMethodException("Unknown property '" + name + "'");


            Method writeMethod = getWriteMethod(descriptor);
            if (writeMethod == null) {
                throw new NoSuchMethodException("Property '" + name + "' has no setter method");
            } else {


                Object values[] = new Object[1];
                values[0] = value;
                writeMethod.invoke(bean, values);

                return;
            }
        }

        public static Method getWriteMethod(PropertyDescriptor descriptor) {
            return getAccessibleMethod(descriptor.getWriteMethod());
        }


}